/****************************************************************************
 *
 * FString1: Klasse FarbString abgeleitet von Klasse String
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 4.3: Vererbung von Klassen mit dynamischen Komponenten
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	String.h
 *
 ****************************************************************************/

#ifndef _String_h
#define _String_h

// Header-Datei fuer I/O
#include <iostream.h>

class String {
  protected:
    char*    cstring;    // Zeichenfolge als dynamischer C-String
    unsigned len;        // Anzahl Zeichen (ohne '\0')
    unsigned size;       // Speicherplatzgroesse von "cstring"

  public:
    // Default- und char*-Konstruktor
    String (const char* = "");

    // Vergleichen von Strings
    friend bool operator== (const String&, const String&);
    friend bool operator!= (const String&, const String&);

    // Hintereinanderhaengen von Strings
    friend String operator+ (const String&, const String&);

    // Ein- und Ausgabe mit Streams
    virtual void printOn (ostream& = cout) const;
    virtual void scanFrom (istream& = cin);

    // Aufgrund dynamischer Komponenten
    String (const String&);                   // Copy-Konstruktor
    const String& operator= (const String&);  // Zuweisung
    virtual ~String ();                       // virtueller Destruktor

  private:
    /* Konstruktor fuer uninitialisierten String bestimmter Laenge
     * (intern fuer operator+())
     */
    String (unsigned);
};

// Standard-Ausgabeoperator
inline ostream& operator<< (ostream& strm, const String& s)
{
    s.printOn (strm);    // String auf Stream ausgeben
    return strm;         // Stream zurueckliefern
}

// Standard-Eingabeoperator
inline istream& operator>> (istream& strm, String& s)
{
    s.scanFrom (strm);   // String von Stream einlesen
    return strm;         // Stream zurueckliefern
}

/* Operator !=
 *  - als Umsetzung auf Operator == inline implementiert
 */
inline bool operator!= (const String& s1, const String& s2) {
    return !(s1==s2);
}

#endif  /* _String_h */


/****************************************************************************
 *
 *	String.cc / String.cpp
 *
 ****************************************************************************/

// Header-Datei der eigenen Klasse
//#include "String.h"

// C-Header-Dateien fuer String-Funktionen
#include <string.h>
#include <ctype.h>

/* Konstruktor aus C-String (char*)
 *  - Default fuer s: ""
 */

String::String (const char* s)
{
    len = strlen (s);             // Anzahl Zeichen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s);          // Zeichen in Speicherplatz kopieren
}

/* Copy-Konstruktor
 */
String::String (const String& s)
{
    len = s.len;                    // Anzahl Zeichen uebernehmen
    size = len + 1;                 // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];      // Speicherplatz anlegen
    strcpy (cstring, s.cstring);    // Zeichen kopieren
}

/* Destruktor
 */
String::~String ()
{
    // Mit new angelegten Speicherplatz wieder freigeben
    delete [] cstring;
}

/* operator =
 *  - Zuweisung
 */
const String& String::operator= (const String& s)
{
    // Zuweisung eines Strings an sich selbst macht nichts
    if (this == &s) {
        return *this;                 // String zurueckliefern    
    }

    len = s.len;             // Anzahl Zeichen uebernehmen

    // falls Platz nicht reicht, vergroessern
    if (size < len + 1) {
        delete [] cstring;            // alten Speicherplatz freigeben
        size = len + 1;               // Anzahl Zeichen plus '\0'
        cstring = new char [size];    // Speicherplatz anlegen
    }

    strcpy (cstring, s.cstring);      // Zeichen kopieren

    return *this;            // geaenderten String zurueckliefern    
}

/* operator ==
 *  - vergleicht zwei Strings
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
bool operator== (const String& s1, const String& s2)
{
    return strcmp(s1.cstring,s2.cstring) == 0;
}

/* operator +
 *  - haengt zwei Strings hintereinander
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
String operator+ (const String& s1, const String& s2)
{
    // uninitialisierten Summen-String erzeugen
    String sum (s1.len + s2.len + 1);

    // Summen-Zeichenfolge darin initialisieren
    strcpy (sum.cstring, s1.cstring);
    strcat (sum.cstring, s2.cstring);

    // Kopie von Summen-String zurueck
    return sum;
}

/* Konstruktor fuer uninitialisierten String bestimmter Laenge
 *  - intern fuer operator+
 */
String::String (unsigned l)
{
    len = l;                      // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
}

/* Ausgabe auf Stream
 */
void String::printOn (ostream& strm) const
{
    // Zeichenfolge einfach ausgeben
    strm << cstring;
}

void String::scanFrom (istream& strm)
{
    char c;

    len = 0;         // zunaechst ist der gelesene String leer

    strm >> ws;      // fuehrende Trennzeichen ueberlesen

    /* solange Input-Stream "strm" nach Einlesen
     * eines Zeichens in "c" in Ordnung ist
     */
    while (strm.get(c)) {       // >> wuerde Trennzeichen ueberlesen

        /* Newline schliesst Stringeingabe ab
         *  - '\n' => RETURN
         */
        if (c == '\n') {
            cstring [len] = '\0';
            return;
        }

        /* falls der Platz nicht mehr ausreicht,
         * muss mehr geschaffen werden
         */
        if (len+1 >= size) {
            char* tmp = cstring;       // Zeiger auf alten Speicherplatz
            size += 80;                // 80 Zeichen mehr Platz
            cstring = new char[size];  // neuen Speicherplatz anlegen
            strcpy (cstring, tmp);     // Zeichen kopieren
            delete [] tmp;             // alten Speicherplatz freigeben
        }

        // neues Zeichen eintragen
        cstring [len] = c;
        len ++;
    }

    /* Einlese-Ende durch Fehler oder EOF
     *  - Stringende-Kennzeichen eintragen
     */
    cstring [len] = '\0';
}



/****************************************************************************
 *
 *	FString.h
 *
 ****************************************************************************/

#ifndef _FarbString_h
#define _FarbString_h

// Header-Datei der Basisklasse
//#include "String.h"

/* Klasse FarbString
 *  - abgeleitet von String
 */
class FarbString : public String {
  protected:
    String farb;    // Farbe, die der String besitzt

  public:
    // Default-, String- und String/String-Konstruktoren
    FarbString (const String& s = "", const String& f = "black")
        : String(s), farb(f) {
    }

    // Farbe abfragen/setzen
    const String& farbe () {
        return farb;
    }
    void farbe (const String& neueFarbe) {
        farb = neueFarbe;
    }

    // Ein- und Ausgabe mit Streams
    virtual void printOn (ostream& = cout) const;
    virtual void scanFrom (istream& = cin);

    // Vergleichen von FarbStrings
    friend bool operator== (const FarbString& s1, const FarbString& s2) {
        return (String)s1 == (String)s2 && s1.farb == s2.farb;
    }
    friend bool operator!= (const FarbString& s1, const FarbString& s2) {
        return !(s1==s2);
    }

    /* Aufgrund dynamischer Komponenten in der Basisklasse
     *  - Copy-Konstruktor und Zuweisung sicherheitshalber,
     *    da nicht klar ist, ob alle Default-Implementierungen
     *    korrekt arbeiten
     */
    FarbString (const FarbString& s)                  // Copy-Konstruktor
        : String(s), farb(s.farb) {
    }
    const FarbString& operator= (const FarbString&);  // Zuweisung

    virtual ~FarbString () {                     // virtueller Destruktor
    }
};

#endif


/****************************************************************************
 *
 *	FString.cc / FString.cpp
 *
 ****************************************************************************/

// Header-Datei der eigenen Klasse
//#include "FarbString.h"

const FarbString& FarbString::operator= (const FarbString& s)
{
    // Zuweisung eines FarbStrings an sich selbst macht nichts
    if (this == &s) {
        return *this;                 // FarbString zurueckliefern    
    }

    // Zuweisungen der Basisklasse durchfuehren
    String::operator= (s);

    // Neue Komponenten (Farbe) zuweisen
    farb = s.farb;

    return *this;            // geaenderten FarbString zurueckliefern    
}

/* Ausgabe auf Stream
 */
void FarbString::printOn (ostream& strm) const
{
    // Zeichenfolge mit Farbe in Klammern ausgeben
    strm << cstring << " (" << farb << ')';
}

/* Einlesen eines FarbStrings von einem Input-Stream
 */
void FarbString::scanFrom (istream& strm)
{
    // Inhalt und Farbe nacheinander einlesen
    String::scanFrom (strm);
    farb.scanFrom (strm);
}


/****************************************************************************
 *
 *	fstest1.cc / fstest1.cpp
 *
 ****************************************************************************/

// Header-Datei fuer I/O
#include <iostream.h>

// Header-Datei fuer Klasse FarbString
//#include "FarbString.h"

main ()
{
    FarbString f ("hallo");
    FarbString r ("rot", "red");

    cout << f << " " << r << endl;    // FarbStrings ausgeben

    f.farbe ("green");                // Farbe von f auf "green" setzen

    cout << f << " " << r << endl;    // FarbStrings ausgeben

    cout << "Summenstring: " << f + r << endl;
}

